"""File with constants used in the SDK."""

# Constant used to add extra timeout to token calculation.
# If the token expiration is lower than this constant, request a new token.
TOKEN_EXTRA_TIME = 10

# REST API only accepts 1MB payload
REST_API_MAXIMUM_CHUNK_SIZE = 1048576



"""File with version and history information."""

__version__ = '1.1.0'

"""
History Information:
1.0.0
    First Version
1.0.1
    Removed tests folder from pip
1.0.2
    Fixed documentation
1.0.3
    Fixed documentation
1.0.4
    Fixed issue #4
1.1.0
    Corrected version number.
    Previous should have been 1.1.0 as it changes function behaviour.
1.2.0
    Moved debug functionality to a separated file.
    Debug output now includes curl command.
"""



# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
from typing import Union
import requests
import time
import urllib3
import pathlib
import urllib

# Internal Imports
# Import only with "from x import y", to simplify the code.

from .common.restobject import RESTObject
from .common.exceptions import RESTAPIError
from .common.exceptions import InvalidOptionError
from .common.constants import TOKEN_EXTRA_TIME
from .common.constants import REST_API_MAXIMUM_CHUNK_SIZE
from .version import __version__

# Disable urllib3 SSL warnings
urllib3.disable_warnings()


class BIG:
    """
    Defines methods to call the iControl REST API that can be used
    by BIG-IP or BIG-IQ.

    Arguments:
        device: Name or IP of the device to send the HTTP requests.
        username: Username used to login to the device.
        password: Password used to login to the device.
        login_provider: Login provider used to authenticate the user.
        request_token: Indicates if a token should be requested from the
            device and used for HTTP requests.
        token: Token to be used to send HTTP requests to the device.
        refresh_token: Refresh token to be used to request new token,
            the new token will be then used for HTTP requests.
        debug: Debug file name to be used to output the debug information.

    Exceptions:
        InvalidOptionError: Raised when invalid options are used as arguments.
    """

    def __init__(self, device: str, username: str = None,
                 password: str = None, login_provider: str = "tmos",
                 request_token: bool = False, token: str = None,
                 refresh_token: str = None, debug: str = None) -> BIG:

        # Input validations
        message = (
            "Use only one option,"
            "request_token or token or refresh_token.")
        if request_token and token is not None:
            raise InvalidOptionError(message)
        if request_token and refresh_token is not None:
            raise InvalidOptionError(message)
        if token is not None and refresh_token is not None:
            raise InvalidOptionError(message)

        # Variables
        self.device = device
        self.username = username
        self.password = password
        self.request_token = request_token
        self.login_provider = login_provider
        self._token = None
        self._token_timeout = None
        self._token_counter = None
        self.token = token
        self.refresh_token = refresh_token
        self._transaction = None
        self.debug = debug
        self.session = requests.Session()

        # Session settings
        if self.request_token is False:
            self.session.auth = (username, password)
        if self.token is not None:
            self.session.headers.update({"X-F5-Auth-Token": f"{self.token}"})
        self.session.headers.update({"Content-Type": "application/json"})
        self.session.verify = False
        self.session.headers.update({"User-Agent": f"BIGREST/{__version__}"})

        # Connect to device
        self._connect()

    def load(self, path: str) -> Union[list[RESTObject], RESTObject]:
        """
        Loads one object or a list of objects from the device.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        response_json = response.json()
        if "items" in response_json:
            collection = []
            for obj in response_json["items"]:
                collection.append(RESTObject(obj))
            return collection
        else:
            return RESTObject(response_json)

    def save(self, obj: RESTObject) -> RESTObject:
        """
        Saves on the device the changes that were made locally to the object.

        Sends an HTTP PUT request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        response = self.session.put(url, json=obj.asdict())
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def delete(self, path: str) -> None:
        """
        Deletes the object on the device.

        Sends an HTTP DELETE request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.delete(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)

    def create(self, path: str, data: dict) -> RESTObject:
        """
        Creates the object on the device.

        Sends an HTTP POST request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            data: Payload that will be sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.post(url, json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def modify(self, path: str, data: dict) -> RESTObject:
        """
        Modifies the object on the device.

        Sends an HTTP PATCH request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            data: Payload that will be sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.patch(url, json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def show(self, path: str) -> list[RESTObject]:
        """
        Gets statistical information about the object from the device.

        Similar to tmsh show command.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        url = f"{url}/stats"
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        response_json = response.json()
        objects = []
        if "entries" in response_json:
            entries = response_json["entries"]
            for obj_name in entries:
                if "nestedStats" in entries[obj_name]:
                    objects.append(
                        RESTObject(
                            entries[obj_name]["nestedStats"]["entries"]))
                else:
                    objects.append(RESTObject(entries))
        else:
            objects.append(RESTObject(response_json))
        return objects

    def command(self, path: str, data: dict) -> str:
        """
        Executes a command on the device.

        Sends an HTTP POST request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            data: Payload that will be sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        obj = self.create(path, data)
        if "commandResult" in obj.properties:
            return obj.properties["commandResult"]
        else:
            return str()

    def download(self, path: str, filename: str) -> None:
        """
        Downloads a file from the device.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            filename: Name of the file to be downloaded.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        # Content-Range: <range-start>-<range-end>/<size>
        if self.request_token or self.refresh_token is not None:
            self._check_token()
        filename_without_path = pathlib.PurePath(filename).name
        url = self._get_url(path)
        url = f"{url}/{filename_without_path}"
        self.session.headers.update(
            {"Content-Type": "application/octet-stream"})
        range_size = REST_API_MAXIMUM_CHUNK_SIZE - 1
        self.session.headers.update(
            {"Content-Range": f"0-{range_size}/0"})
        response = self.session.get(url)
        if response.status_code != 200 and response.status_code != 206:
            raise RESTAPIError(self.esponse, self.debug)
        content_range = response.headers.get("Content-Range")
        range_start = range_size + 1
        range_end = range_start + range_size
        size = int(content_range.split("/")[1])
        size = size - 1
        with open(filename, "wb") as file_:
            file_.write(response.content)
            while size >= range_start:
                if range_end > size:
                    range_end = size
                content_range = f"{range_start}-{range_end}/{size + 1}"
                self.session.headers.update(
                    {"Content-Range": content_range})
                response = self.session.get(url)
                file_.write(response.content)
                if response.status_code != 200:
                    raise RESTAPIError(response, self.debug)
                range_start = range_end + 1
                range_end = range_end + range_size + 1
        self.session.headers.pop("Content-Range")
        self.session.headers.update({"Content-Type": "application/json"})

    def upload(self, path: str, filename: str) -> None:
        """
        Uploads a file to the device.

        Sends an HTTP POST request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            filename: Name of the file to be uploaded.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        # Content-Range: <range-start>-<range-end>/<size>
        if self.request_token or self.refresh_token is not None:
            self._check_token()
        self.session.headers.update(
            {"Content-Type": "application/octet-stream"})
        filename_without_path = pathlib.PurePath(filename).name
        url = self._get_url(path)
        url = f"{url}/{filename_without_path}"
        range_start = 0
        range_size = REST_API_MAXIMUM_CHUNK_SIZE - 1
        range_end = range_size
        size = pathlib.Path(filename).stat().st_size - 1
        with open(filename, "rb") as file_:
            if size <= range_size:
                self.session.headers.update(
                    {"Content-Range": f"{range_start}-{size}/{size + 1}"})
                response = self.session.post(url, data=file_.read(size + 1))
                if response.status_code != 200:
                    raise RESTAPIError(response, self.debug)
            else:
                while size >= range_start:
                    if range_end > size:
                        range_end = size
                    content_range = f"{range_start}-{range_end}/{size + 1}"
                    self.session.headers.update(
                        {"Content-Range": content_range})
                    bytes_to_read = (range_end - range_start) + 1
                    response = self.session.post(
                        url, data=file_.read(bytes_to_read))
                    if response.status_code != 200:
                        raise RESTAPIError(response, self.debug)
                    range_start = range_end + 1
                    range_end = range_end + range_size + 1
        raise RESTAPIError(response, self.debug)
        self.session.headers.pop("Content-Range")
        self.session.headers.update({"Content-Type": "application/json"})

    def example(self, path: str) -> RESTObject:
        """
        Gets an example of an object from the device.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        url = f"{url}/example"
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def _get_url(self, path: str) -> str:
        """
        Creates the URL to be used to connect to the device.

        Arguments:
            path: HTTP path used to create the URL.
        """

        if path.endswith("/"):
            path = path[:-1]
        return f"https://{self.device}{path}"

    def _check_token(self) -> None:
        """
        Verifies if there is a token to be used.

        If there is a token, it verifies if it is still valid.
        """

        if self._token is None:
            self._get_token()
        else:
            time_passed = time.time() - self._token_counter
            # Only use token if still valid for TOKEN_EXTRA_TIME seconds
            time_passed = time_passed + TOKEN_EXTRA_TIME
            if time_passed > self._token_timeout:
                self._get_token()

    def _get_token(self) -> None:
        """
        Gets a token from the device.

        Sends an HTTP POST request to the iControl REST API.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if "X-F5-Auth-Token" in self.session.headers:
            self.session.headers.pop("X-F5-Auth-Token")
        # Required for 12.0.0, but not in 15.1.0
        self.session.auth = (self.username, self.password)
        response = None
        if self.refresh_token is None:
            data = {}
            data["username"] = self.username
            data["password"] = self.password
            data["loginProviderName"] = self.login_provider
            response = self.session.post(
                f"https://{self.device}/mgmt/shared/authn/login", json=data)
        else:
            data_token = {}
            data_token["token"] = self.refresh_token
            data = {}
            data["refreshToken"] = data_token
            response = self.session.post(
                f"https://{self.device}/mgmt/shared/authn/exchange", json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        self._token_counter = time.time()
        self._token = response.json()["token"]["token"]
        self.session.headers.update({"X-F5-Auth-Token": f"{self._token}"})
        self._token_timeout = response.json()["token"]["timeout"]
        self.session.auth = None

    def _get_path(self, obj: RESTObject) -> str:
        """
        Gets the HTTP path from the object.

        Arguments:
            obj: Object to get the path from.
        """
        self_link = obj.properties["selfLink"]
        path = urllib.parse.urlparse(self_link).path
        return path

    def _connect(self) -> None:
        """
        Test a connection to the device.

        If it is not using basic authentication, tries to get a token.

        if it is using basic authentication, sends an HTTP POST request
            to the iControl REST API.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._get_token()
        elif self.token is not None:
            response = self.session.get(
                f"https://{self.device}/mgmt/shared/echo-query")
            if response.status_code != 200:
                raise RESTAPIError(response, self.debug)
        else:
            data = {}
            data["username"] = self.username
            data["password"] = self.password
            data["loginProviderName"] = self.login_provider
            response = self.session.post(
                f"https://{self.device}/mgmt/shared/authn/login", json=data)
            if response.status_code != 200:
                raise RESTAPIError(response, self.debug)



# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import time

# Internal Imports
# Import only with "from x import y", to simplify the code.

from .big import BIG
from .common.exceptions import RESTAPIError
from .common.restobject import RESTObject


class BIGIP(BIG):
    """
    Defines methods to call the iControl REST API that can be used by BIG-IP.

    Arguments:
        device: Name or IP of the device to send the HTTP requests.
        username: Username used to login to the device.
        password: Password used to login to the device.
        login_provider: Login provider used to authenticate the user.
        request_token: Indicates if token a should be requested from the
            device and used for HTTP requests.
        token: Token to be used to send HTTP requests to the device.
        debug: Debug file name to be used to output the debug information.

    Exceptions:
        InvalidOptionError: Raised when invalid options are used as arguments.
    """

    def __init__(self, device: str, username: str = None,
                 password: str = None, login_provider: str = "tmos",
                 request_token: bool = False, token: str = None,
                 debug: str = None) -> BIGIP:
        super().__init__(
            device=device, username=username, password=password,
            login_provider=login_provider, request_token=request_token,
            token=token, debug=debug)

    def task_start(self, path: str, data: dict) -> RESTObject:
        """
        Starts a task on the device.

        Sends HTTP POST and PUT requests to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            data: Payload that will be sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.post(url, json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        id_ = response.json()["_taskId"]
        data = {}
        data["_taskState"] = "VALIDATING"
        url = f"{url}/{id_}"
        response_put = self.session.put(url, json=data)
        if response_put.status_code != 202:
            raise RESTAPIError(response_put)
        return RESTObject(response.json())

    def task_wait(self, obj: RESTObject, interval: int = 10) -> None:
        """
        Continually queries the status of the task until it finishes.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            obj: Object that represents the task.
            interval: The interval the queries will be made.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        while True:
            if self.request_token or self.refresh_token is not None:
                self._check_token()
            response = self.session.get(url)
            if response.status_code != 200:
                raise RESTAPIError(response, self.debug)
            status = response.json()["_taskState"]
            if status == "FAILED":
                raise RESTAPIError(response, self.debug)
            if status == "COMPLETED":
                return
            else:
                time.sleep(interval)

    def task_completed(self, obj: RESTObject) -> bool:
        """
        Verifies if the task is completed.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            obj: Object that represents the task.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        status = response.json()["_taskState"]
        if status == "FAILED":
            raise RESTAPIError(response, self.debug)
        if status == "COMPLETED":
            return True
        else:
            return False

    def task_result(self, obj: RESTObject) -> str:
        """
        Get the result text of the task.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            obj: Object that represents the task.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        url = f"{url}/result"
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        if "commandResult" in response.json():
            return response.json()["commandResult"]
        else:
            return str()

    def exist(self, path: str) -> bool:
        """
        Verifies if the object exists on the device.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.get(url)
        if response.status_code == 200:
            return True
        if response.status_code == 404:
            return False
        raise RESTAPIError(response, self.debug)

    def transaction_create(self) -> RESTObject:
        """
        Create a transaction on the device.

        Sends an HTTP POST request to the iControl REST API.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url("/mgmt/tm/transaction")
        response = self.session.post(url, json={})
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        self._transaction = response.json()["transId"]
        self.session.headers.update(
            {"X-F5-REST-Coordination-Id": f"{self._transaction}"})
        return RESTObject(response.json())

    def __enter__(self) -> RESTObject:
        """
        Used in with statement.

        Calls transaction_create().
        """

        return self.transaction_create()

    def transaction_commit(self) -> RESTObject:
        """
        Commit a transaction on the device.

        Sends an HTTP PATCH request to the iControl REST API.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(f"/mgmt/tm/transaction/{self._transaction}")
        self.session.headers.pop("X-F5-REST-Coordination-Id")
        data = {}
        data["state"] = "VALIDATING"
        response = self.session.patch(url, json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def __exit__(self, type, value, traceback) -> None:
        """
        Used in with statement.

        Calls transaction_commit().
        """

        self.transaction_commit()

    def transaction_validate(self) -> RESTObject:
        """
        Validate a transaction on the device.

        Sends an HTTP PATCH request to the iControl REST API.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(f"/mgmt/tm/transaction/{self._transaction}")
        self.session.headers.pop("X-F5-REST-Coordination-Id")
        data = {}
        data["validateOnly"] = True
        data["state"] = "VALIDATING"
        response = self.session.patch(url, json=data)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        else:
            self.session.headers.update(
                {"X-F5-REST-Coordination-Id": f"{self._transaction}"})
        return RESTObject(response.json())



"""Defines a list of exceptions."""

# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import requests
import json

# Internal Imports
# Import only with "from x import y", to simplify the code.
from .debug import debug_data
from .debug import debug_curl


class Error(Exception):
    """Base class for all SDK exceptions."""
    pass


class RESTAPIError(Error):
    """
    Exception for API errors from iControl REST API.

    Arguments:
        response: A object created from a response to a HTTP request.
        debug: If set, indicates the file to save the debug output created.
    """

    def __init__(self, response: requests.Response,
                 debug: str) -> RESTAPIError:
        if debug is not None:
            with open(debug, "a") as file_:
                file_.write(debug_data(response))
                file_.write(debug_curl(response))
                super().__init__(f"Debug information saved in file {debug}.")
        else:
            reponse_content_type = response.headers.get("Content-Type")
            try:
                response_json = json.dumps(response.json(), indent=4)
                response_body = f"\nResponse Body:\n{response_json}"
            except Exception:
                if reponse_content_type == "application/octet-stream":
                    response_body = f"\nResponse Body:\n<binary data>"
                else:
                    response_body = f"\nResponse Body:\n{response.text}"
            response_status = response.status_code
            information = f"\nStatus:\n{response_status}"
            super().__init__(f"{information}{response_body}")


class InvalidOptionError(Error):
    """
    Exception raised for invalid options passed.

    Arguments:
        message: Message that will be printed with the exception.
    """
    def __init__(self, message: str) -> InvalidOptionError:
        super().__init__(message)



import setuptools
from bigrest.version import __version__

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bigrest",
    version=__version__,
    author="Leonardo Souza",
    description="A Python SDK for F5 iControl REST API",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/leonardobdes/BIGREST",
    packages=setuptools.find_packages(exclude=("tests",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



"""Defines functions to be use for debug."""

# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import requests
import json


def debug_data(response: requests.Response) -> str:
    """
    Provides information about the HTTP request
    and HTTP response to be used for troubleshooting.

    Arguments:
        response: A object created from a response to a HTTP request.
    """

    request_content_type = response.request.headers.get("Content-Type")
    reponse_content_type = response.headers.get("Content-Type")
    try:
        response_body = json.dumps(response.json(), indent=4)
    except Exception:
        if reponse_content_type == "application/octet-stream":
            response_body = "<binary data>"
        else:
            response_body = response.text
    response_status = response.status_code
    request_method = response.request.method
    request_url = response.request.url
    request_headers = response.request.headers
    request_headers = json.dumps(dict(request_headers), indent=4)
    request_body = response.request.body
    if request_body is not None:
        try:
            request_body = json.dumps(
                dict(eval(request_body)), indent=4)
        except Exception:
            if request_content_type == "application/octet-stream":
                request_body = "<binary data>"
            else:
                request_body = response.text
    else:
        request_body = ""
    response_headers = response.headers
    response_headers = json.dumps(dict(response_headers), indent=4)
    information = (f"Request Method:\n{request_method}"
                   f"\nRequest URL:\n{request_url}"
                   f"\nRequest Headers:\n{request_headers}"
                   f"\nRequest Body:\n{request_body}"
                   f"\nResponse Status:\n{response_status}"
                   f"\nResponse Headers:\n{response_headers}"
                   f"\nResponse Body:\n{response_body}\n")
    return information


def debug_curl(response: requests.Response) -> str:
    """
    Converts a HTTP to a curl format to be used for troubleshooting.

    Arguments:
        response: A object created from a response to a HTTP request.
    """

    request = f"Curl Command:\n"
    request += f"curl -k -X {response.request.method} {response.request.url}"
    headers = response.request.headers
    for header in headers:
        request += f" -H '{header}: {headers[header]}'"
    body = response.request.body
    if body is not None:
        try:
            json.dumps(body)
            request += f" -d '{body}'"
        except Exception:
            request += f" --data-binary @filename"
    return f"{request}\n"
